#include "error.hh"

#include <cstring>

/* This table of error messages comes almost directly
 * from QuickBASIC. Most of the messages are obsolete,
 * but they're maintained for compatibility with older
 * BASIC versions such as GW-BASIC.
 * (!) indicates that the message does not exist in QB.
 */
static const char* const ErrorMessages[] =
{
    "NEXT without FOR", //1,
    "Syntax error", //2,
    "RETURN without GOSUB", //3,
    "Out of DATA", //4,
    "Illegal function call", //5,
    "Overflow", //6,
    "Out of memory", //7,
    "Label not defined", //8,
    "Subscript out of range", //9,
    "Duplicate definition", //10,
    "Division by zero", //11,
    "Illegal in direct mode", //12,
    "Type mismatch", //13,
    "Out of string space", //14,
    "String too long", // 15 (!)
    "String formula too complex", //16,
    "Cannot continue", //17,
    "Function not defined", //18,
    "No RESUME", //19,
    "RESUME without error", //20,
    "Unprintable error", // 21
    "Missing operand", // 22 (!)
    "Line buffer overflow", // 23 (!)
    "Device timeout", //24,
    "Device fault", //25,
    "FOR without NEXT", //26,
    "Out of paper", //27,
    0,
    "WHILE without WEND", //29,
    "WEND without WHILE", //30,
    0,
    0,
    "Duplicate label", //33,
    0,
    "Subprogram not defined", //35,
    "Subprogram already in use", //36 (!),
    "Argument-count mismatch", //37,
    "Array not defined", //38,
    "CASE ELSE expected", //39,
    "Variable required", //40,
    0,0,0,0,
    0,0,0,0,
    0,
    "FIELD overflow", //50,
    "Internal error", //51,
    "Bad file name or number", //52,
    "File not found", //53,
    "Bad file mode", //54,
    "File already open", //55,
    "FIELD statement active", //56,
    "Device I/O error", //57,
    "File already exists", //58,
    "Bad record length", //59,
    0,
    "Disk full", //61,
    "Input past end of file", //62,
    "Bad record number", //63,
    "Bad file name", //64,
    0,
    "Direct statement in file", //66,
    "Too many files", //67,
    "Device unavailable", //68,
    "Communication-buffer overflow", //69,
    "Permission denied", //70,
    "Disk not ready", //71,
    "Disk-media error", //72,
    "Advanced feature unavailable", //73,
    "Rename across disks", //74,
    "Path/File access error", //75,
    "Path not found", //76,
    "Deadlock" //77,
};

size_t HandlingError = 0;
std::vector<jmp_buf_wrapper> ErrorFrameStack;
jmp_buf_wrapper ErrorSpot, SafeReturnToMain;
void* ErrorHandler = 0, *ResumeTarget = 0;
int B_ERL_VALUE, B_ERR_VALUE;

void B_RESUME_FAILTEST()
{
    if(!HandlingError) B_ERROR(20);/* RESUME WITHOUT ERROR*/
    if(ErrorFrameStack.empty())
    {
        fprintf(stderr, "Warning: Source compiled without RESUME support\n");
    }
}
/* RESUME and RESUMENEXT may return into a subprogram, but also to main program */
void B_RESUME()
{
    B_RESUME_FAILTEST();
    
    // peels off the frames possible created by the error handler
    ErrorFrameStack.resize(HandlingError);
    
    longjmp(ErrorFrameStack.back(), B_ERR_RESUME_REPEAT_CODE);
}
void B_RESUME_NEXT()
{
    B_RESUME_FAILTEST();
    ErrorFrameStack.resize(HandlingError);
    longjmp(ErrorFrameStack.back(), B_ERR_RESUME_NEXT_CODE);
}
void B_ERR_PUSH()
{
    ErrorFrameStack.push_back(ErrorSpot);
}
void B_ERR_POP()
{
    ErrorSpot = ErrorFrameStack.back();
    ErrorFrameStack.pop_back();
}
void B_ERROR(BAS_INT code)
{
    B_ERL_VALUE = 0; //FIXME
    B_ERR_VALUE = code;
    
    if(!ErrorHandler)
    {
        const BAS_INT NumMessages = sizeof(ErrorMessages) / sizeof(*ErrorMessages);
        int print_code = code;
        if(code < 1
        || code > NumMessages
        || !ErrorMessages[code-1]) print_code = 21; // Unprintable error
        fprintf(stderr, "%s (error %d)\n", ErrorMessages[print_code-1], (int)code);
        exit(code);
    }
    HandlingError = ErrorFrameStack.size();
    goto *ErrorHandler;
}
