#include "expr.hh"
#include "dimdef.hh"
#include "casevalue.hh"
#include "binding.hh"

#include "lexer.hh"
#include "tokens.hh"

//#define BOOST_SPIRIT_DEBUG        // define this for debug output

#if 1
#define BOOST_SPIRIT_SWITCH_CASE_LIMIT 70

#include <boost/spirit/core.hpp>
#include <boost/spirit/utility/loops.hpp>
#include <boost/spirit/dynamic/if.hpp>
#include <boost/spirit/dynamic/for.hpp>
#include <boost/spirit/dynamic/while.hpp>
#include <boost/spirit/dynamic/switch.hpp>
#include <boost/spirit/iterator/multi_pass.hpp>
#include <boost/spirit/actor/ref_const_ref_const_ref_a.hpp>
#include <boost/spirit/actor/assign_actor.hpp>
#include <boost/spirit/actor/increment_actor.hpp>
#else

#include "parser.hh"

#endif

/*
namespace boost { namespace spirit { namespace impl {
static bool iscntrl_(int c) { return false; }
} } }
*/

namespace boost {
namespace spirit {
namespace multi_pass_policies {
    int yylex()
    {
        QBtoken tok = (QBtoken)YY_LEXFUNAME();
        fprintf(stderr, "Returning token: %s\n", GetTokenName(tok).c_str());    
        return tok;
    }
}
} }

using namespace boost::spirit;

#define tk(x) ch_p((QBtoken)x)

/*
template<class T>
struct replace_new_tok_action
{
    template<typename T1,typename T2,typename T3,typename T4>
    void act(T1&a, T2&b, T3&c, T4&d)
    {
        a = new T(a, b, c, d);
    }
    template<typename T1,typename T2,typename T3>
    void act(T1&a, T2&b, T3&c)
    {
        a = new T(a, b, c);
    }
    template<typename T1,typename T2>
    void act(T1&a, T2&b)
    {
        a = new T(a, b);
    }
};

template<class T, typename T1,typename T2,typename T3>
inline ref_const_ref_const_ref_actor<T1,T2,T3, replace_new_tok_action<T> >
replace_new_tok(T1&a,T2&b,T3&c)
{
    return ref_const_ref_const_ref_actor<T1,T2,T3,replace_new_tok_action<T> >
            (a,b,c);
};
template<class T, typename T1,typename T2>
inline ref_const_ref_actor<T1,T2, replace_new_tok_action<T> >
replace_new_tok(T1&a,T2&b)
{
    return ref_const_ref_actor<T1,T2,replace_new_tok_action<T> >
            (a,b);
};
*/

namespace
{
    exprp exp_lhs;
    exprp exp_result;
    unsigned notcount;
    QBtoken matched_token;
    
    void SetExpr(exprp new_res)
    {
        exp_lhs    = exp_result;
        exp_result = new_res;
    }
    
    template<typename E>
    struct DoBinaryExpr
    {
        void act(int op) const
        {
            SetExpr(new E(op, exp_lhs, exp_result));
        }
    };
    
    template<typename E, int op>
    struct DoUnaryToggleExpr
    {
        void act(int count) const
        {
            if(count % 2)
                SetExpr(new E(op, exp_result));
        }
    };
    
    template<typename E>
    struct DoUnaryExpr
    {
        void act(int op) const
        {
            SetExpr(new E(op, exp_result));
        }
    };
    
    template<typename It>
    void DoNumber(const It&,const It&)
    {
        SetExpr(new expr_number(yylval.number));
    }
    template<typename It>
    void DoString(const It&,const It&)
    {
        SetExpr(new expr_string(yylval.str));
    }
    
    template<typename T, typename ActionT>
    class int_actor: public ActionT
    {
        T val;
    public:
        explicit int_actor(const T& v) : val(v) { }
    
        template<typename It>
        void operator() (const It&,const It&) const
        {
            this->act(val);
        }
        template<typename It>
        void operator() () const
        {
            this->act(val);
        }
    };
    
    void CounterReset() { notcount=0; }
    void CounterInc() { ++notcount; }
};

static struct QB_rvalue_parser : public grammar<QB_rvalue_parser>
{
    template<typename ScannerT>
    struct result
    {
        typedef typename match_result<ScannerT, exprp>::type type;
    };
    
    template<typename ScannerT>
    struct definition
    {
        #define tk(x) ch_p((QBtoken)x)
        
        definition(const QB_rvalue_parser& )
        {
            first =
            (
                #define binary_op_while(subclass, op) \
                    while_p(tk(op)) [ subclass [ int_actor<int,DoBinaryExpr<expr_binary> > (op) ] ]
            
                expr     = expr_IMP >> binary_op_while(expr_IMP, IMP),
                expr_IMP = expr_EQV >> binary_op_while(expr_EQV, EQV),
                expr_EQV = expr_XOR >> binary_op_while(expr_XOR, XOR),
                expr_XOR = expr_OR  >> binary_op_while(expr_OR , OR),
                expr_OR  = expr_AND >> binary_op_while(expr_AND, AND),
                expr_AND =
                  for_p(&CounterReset, tk(NOT), &CounterInc) [ epsilon_p ]
                  >> expr_NOT [ int_actor<int,DoUnaryToggleExpr<expr_unary, NOT> > (boost::ref(notcount)) ],
                expr_NOT = expr_Sum >>
                      while_p( ( tk(NOTEQUAL) | tk(LESSOREQUAL)
                               | tk(GREATEROREQUAL) | tk('>')
                               | tk('<') | tk('=') ) [ assign_a(matched_token) ] )
                      [ expr_Sum [ int_actor<int,DoBinaryExpr<expr_binary> > (boost::ref(matched_token)) ] ],
                expr_Sum = expr_MOD  >> binary_op_while(expr_MOD,  MOD ),
                expr_MOD = expr_IDIV >> binary_op_while(expr_IDIV, '\\' ),
                expr_IDIV = expr_Mul >>
                      while_p( ( tk('*') | tk('/') ) [ assign_a(matched_token) ] )
                      [ expr_Mul [ int_actor<int,DoBinaryExpr<expr_binary> > (boost::ref(matched_token)) ] ],
                expr_Mul =
                  for_p(&CounterReset, tk('-'), &CounterInc) [ epsilon_p ]
                  >> expr_Neg [ int_actor<int,DoUnaryToggleExpr<expr_unary, '-'> > (boost::ref(notcount)) ],
                expr_Neg = expr_Pow >> binary_op_while(expr_Mul, '^'),
                
                expr_Pow =
                    tk('(') >> expr >> tk(')')
                |   expr_meat,
                
                expr_meat =
                    switch_p
                    [
                        case_p<ABS>(tk('(') >> num_expr >> tk(')'))     [ int_actor<int,DoUnaryExpr<expr_builtin_numeric> > (ABS) ],
                        case_p<ASC>(tk('(') >> string_expr >> tk(')'))  [ int_actor<int,DoUnaryExpr<expr_builtin_numeric> > (ASC) ],
                        case_p<ATN>(tk('(') >> num_expr >> tk(')')),        /* $$ = new expr_builtin_numeric(ATN, $3); */
                        case_p<CDBL>(tk('(') >> num_expr >> tk(')')),       /* $$ = new expr_builtin_numeric(CDBL, $3); */
                        case_p<CHR_S>(tk('(') >> num_expr >> tk(')')),      /* $$ = new expr_builtin_string(CHR_S, $3); */
                        case_p<CINT>(tk('(') >> num_expr >> tk(')')),       /* $$ = new expr_builtin_numeric(CINT, $3); */
                        case_p<CLNG>(tk('(') >> num_expr >> tk(')')),       /* $$ = new expr_builtin_numeric(CLNG, $3); */
                        case_p<COS>(tk('(') >> num_expr >> tk(')')),        /* $$ = new expr_builtin_numeric(COS, $3); */
                        case_p<CSNG>(tk('(') >> num_expr >> tk(')')),       /* $$ = new expr_builtin_numeric(CSNG, $3); */
                        case_p<CSRLIN>( epsilon_p ),                                       /* $$ = new expr_builtin_numeric(CSRLIN); */
                        case_p<CVD>(tk('(') >> string_expr >> tk(')')),                                   /* $$ = new expr_builtin_numeric(CVD, $3); */
                        case_p<CVI>(tk('(') >> string_expr >> tk(')')),                                   /* $$ = new expr_builtin_numeric(CVI, $3); */
                        case_p<CVL>(tk('(') >> string_expr >> tk(')')),                                   /* $$ = new expr_builtin_numeric(CVL, $3); */
                        case_p<CVS>(tk('(') >> string_expr >> tk(')')),                                   /* $$ = new expr_builtin_numeric(CVS, $3); */
                        case_p<DATE_S>( epsilon_p ),                                       /* $$ = new expr_builtin_string(DATE_S); */
                        case_p<ENVIRON_S>(tk('(') >> string_or_num_expr >> tk(')')),                     /* $$ = new expr_builtin_string(ENVIRON_S, $3); */
                        case_p<ERL>( epsilon_p ),                                          /* $$ = new expr_builtin_numeric(ERL); */
                        case_p<ERR>( epsilon_p ),                                          /* $$ = new expr_builtin_numeric(ERR); */
                        case_p<EXP>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(EXP, $3); */
                        case_p<FILEATTR>(tk('(') >> autofilenum >> tk(',') >> num_expr >> tk(')')),                /* $$ = new expr_builtin_numeric(FILEATTR, $3, $5); */
                        case_p<FIX>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(FIX, $3); */
                        case_p<FRE>(tk('(') >> string_or_num_expr >> tk(')')),                             /* $$ = new expr_builtin_numeric(FRE, $3); */
                        case_p<FREEFILE>( epsilon_p ),                                     /* $$ = new expr_builtin_numeric(FREEFILE); */
                        case_p<HEX_S>(tk('(') >> num_expr >> tk(')')),                                    /* $$ = new expr_builtin_string(HEX_S, $3); */
                        case_p<INKEY_S>( epsilon_p ),                                      /* $$ = new expr_builtin_string(INKEY_S); */
                        case_p<INP>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(INP, $3); */
                        case_p<INPUT_S>(tk('(') >> num_expr >> tk(')')
                                      | tk('(') >> num_expr >> tk(',') >> autofilenum >> tk(')')),        /* $$ = new expr_builtin_string(INPUT_S, $3, $5); */
                        case_p<INT>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(INT, $3); */
                        case_p<INSTR>(
                             tk('(') >> string_expr >> tk(',') >> string_expr >> tk(')')
                           | tk('(') >> num_expr
                          >> tk(',') >> string_expr >> tk(',') >> string_expr >> tk(')')),       /* $$ = new expr_builtin_string(INSTR, $3, $5, $7); */
                        case_p<LBOUND>(tk('(') >> var_target >> if_p(tk(',')) [ num_expr ] >> tk(')')),              /* $$ = new expr_builtin_numeric(LBOUND, $3, $5); */
                        case_p<LCASE_S>(tk('(') >> string_expr >> tk(')')),                               /* $$ = new expr_builtin_string(LCASE_S, $3); */
                        case_p<LEFT_S>(tk('(') >> string_expr >> tk(',') >> num_expr >> tk(')')),                    /* $$ = new expr_builtin_string(LEFT_S, $3, $5); */
                        case_p<LEN>(tk('(') >> /*string or any*/expr >> tk(')')),                                   /* $$ = new expr_builtin_numeric(LEN, $3); */
                        case_p<LOC>(tk('(') >> autofilenum >> tk(')')),                                  /* $$ = new expr_builtin_numeric(LOC, $3); */
                        case_p<LOF>(tk('(') >> autofilenum >> tk(')')),                                  /* $$ = new expr_builtin_numeric(LOF, $3); */
                        case_p<LOG>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(LOG, $3); */
                        case_p<LPOS>(tk('(') >> num_expr >> tk(')')),                                     /* $$ = new expr_builtin_numeric(LPOS, $3); */
                        case_p<LTRIM_S>(tk('(') >> string_expr >> tk(')')),                               /* $$ = new expr_builtin_string(LTRIM_S, $3); */
                        case_p<MID_S>(tk('(') >> string_expr >> tk(',') >> num_expr >> if_p(tk(',')) [ num_expr ] >> tk(')')),  /* $$ = new expr_builtin_string(MID_S, $3, $5, $7); */
                        case_p<MKD_S>(tk('(') >> num_expr >> tk(')')),                                    /* $$ = new expr_builtin_string(MKD_S, $3); */
                        case_p<MKI_S>(tk('(') >> num_expr >> tk(')')),                                    /* $$ = new expr_builtin_string(MKI_S, $3); */
                        case_p<MKL_S>(tk('(') >> num_expr >> tk(')')),                                    /* $$ = new expr_builtin_string(MKL_S, $3); */
                        case_p<MKS_S>(tk('(') >> num_expr >> tk(')')),                                    /* $$ = new expr_builtin_string(MKS_S, $3); */
                        case_p<OCT_S>(tk('(') >> num_expr >> tk(')')),                                    /* $$ = new expr_builtin_string(OCT_S, $3); */
                        case_p<PEEK>(tk('(') >> num_expr >> tk(')')),                                     /* $$ = new expr_builtin_numeric(PEEK, $3); */
                        case_p<PLAY>(tk('(') >> dummy_expr >> tk(')')),                                  /* $$ = new expr_builtin_numeric(PLAY); */
                        case_p<POINT>(tk('(') >> num_expr >> if_p(tk(',')) [ num_expr ] >> tk(')')),                     /* $$ = new expr_builtin_numeric(POINT, $3, $5); */
                        case_p<POS>(tk('(') >> dummy_expr >> tk(')')),                                   /* $$ = new expr_builtin_numeric(POS); */
                        case_p<RIGHT_S>(tk('(') >> string_expr >> tk(',') >> num_expr >> tk(')')),                   /* $$ = new expr_builtin_string(RIGHT_S, $3, $5); */
                        case_p<RND>( if_p(tk('(')) [ num_expr >> tk(')') ] ),                           /* $$ = new expr_builtin_numeric(RND, $3); */
                        case_p<RTRIM_S>(tk('(') >> string_expr >> tk(')')),                               /* $$ = new expr_builtin_string(RTRIM_S, $3); */
                        case_p<SPACE_S>(tk('(') >> num_expr >> tk(')')),                                  /* $$ = new expr_builtin_string(SPACE_S, $3); */
                        case_p<SPC>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(SPC, $3); */
                        case_p<SQR>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(SQR, $3); */
                        case_p<STRING_S>(tk('(') >> num_expr >> tk(',') >> string_or_num_expr >> tk(')')),          /* $$ = new expr_builtin_string(STRING_S, $3, $5); */
                        case_p<STR_S>(tk('(') >> num_expr >> tk(')')),                                    /* $$ = new expr_builtin_string(STR_S, $3); */
                        case_p<TAB>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(TAB, $3); */
                        case_p<TAN>(tk('(') >> num_expr >> tk(')')),                                      /* $$ = new expr_builtin_numeric(TAN, $3); */
                        case_p<TIMER>( epsilon_p ),                                        /* $$ = new expr_builtin_numeric(TIMER); */
                        case_p<TIME_S>( epsilon_p ),                                       /* $$ = new expr_builtin_string(TIME_S); */
                        case_p<UBOUND>(tk('(') >> var_target >> if_p(tk(',')) [ num_expr ] >> tk(')')),              /* $$ = new expr_builtin_numeric(UBOUND, $3, $5); */
                        case_p<UCASE_S>(tk('(') >> string_expr >> tk(')')),                               /* $$ = new expr_builtin_string(UCASE_S, $3); */
                        case_p<VAL>(tk('(') >> string_expr >> tk(')')),                                   /* $$ = new expr_builtin_numeric(VAL, $3); */
                        case_p<VARPTR>(tk('(') >> var_target >> tk(')')),                                /* $$ = new expr_builtin_numeric(VARPTR, $3); */
                        case_p<VARPTR_S>(tk('(') >> var_target >> tk(')')),                              /* $$ = new expr_builtin_string(VARPTR_S, $3); */
                        case_p<VARSEG>(tk('(') >> var_target >> tk(')')),                                /* $$ = new expr_builtin_numeric(VARSEG, $3); */
                        
                        case_p<NUMBER>(      epsilon_p [ DoNumber ] ),
                        case_p<STRINGCONST>( epsilon_p [ DoString ] )
                    ]
                |   var_or_const_or_functioncall,
                
                /**** RVAL */
                var_or_const_or_functioncall =
                    (
                        tk(IDENTIFIER)
                        /* perhaps with indexes */
                        >> if_p(tk('(')) [ (num_expr % tk(',')) >> tk(')') ]
                    ) % tk('.'),
                      /* expr_var* parent = dynamic_cast<expr_var*>($1);
                         expr_var* member = dynamic_cast<expr_var*>($3);
                         parent->set_ptr(member); $3 = 0; $$ = $1;
                       */
                    /* If it had indexes, it can be ARRAY/FUNCTION name.
                     * If it had no indexes, it can be CONST/SCALAR/FUNCTIONNAME.
                     * If it had members, it was an user type.
                     */

                var_target =
                    (
                        tk(IDENTIFIER)
                        /* perhaps with indexes */
                        >> if_p(tk('(')) [ (num_expr % tk(',')) >> tk(')') ]
                    ) % tk('.'),
                      /* expr_var* parent = dynamic_cast<expr_var*>($1);
                         expr_var* member = dynamic_cast<expr_var*>($3);
                         parent->set_ptr(member); $3 = 0; $$ = $1;
                       */

                autofilenum = tk(FILENUM) | num_expr
            );
        }

        subrule<0> expr;
        /* aliases to expr: */
        subrule<0> num_expr;
        subrule<0> bool_expr;
        subrule<0> string_expr;
        subrule<0> string_or_num_expr;
        subrule<0> dummy_expr; /* discardable value */

        subrule<1> expr_IMP;
        subrule<2> expr_EQV;
        subrule<3> expr_XOR;
        subrule<4> expr_OR;
        subrule<5> expr_AND;
        subrule<6> expr_NOT;
        subrule<7> expr_Sum;
        subrule<8> expr_MOD;
        subrule<9> expr_IDIV;
        subrule<10> expr_Mul;
        subrule<11> expr_Neg;
        subrule<12> expr_Pow;
        subrule<13> expr_meat;
        
        subrule<14> var_or_const_or_functioncall;
        subrule<15> var_target;
        subrule<16> autofilenum;
        
        rule<ScannerT> first;
        
        const rule<ScannerT>& start() const { return first; }
    };
} QB_rvalue_p;

struct QB_Parser : public grammar<QB_Parser>
{
    template<typename ScannerT>
    struct definition
    {
        definition(const QB_Parser& )
        {
            first =
            (
                program =
                *(  switch_p
                    [
                        case_p<DECLARESUB> (
                            tk(IDENTIFIER) >> sub_params_decl
                         >> tk(CRLF)),
                        case_p<DECLAREFUNCTION> (
                            tk(IDENTIFIER) >> sub_params_decl
                         >> tk(CRLF)),
                        case_p<SUB> (
                            tk(IDENTIFIER) >> !sub_params_decl
                         >> !tk(STATIC) >> tk(CRLF)
                         >> longstatements0
                         >> tk(ENDSUB) >> tk(CRLF)),
                        case_p<FUNCTION> (
                            tk(IDENTIFIER) >> !sub_params_decl
                         >> !tk(STATIC) >> separator
                         >> longstatements0
                         >> tk(ENDFUNCTION) >> tk(CRLF)),
                        case_p<TYPE> (
                            tk(IDENTIFIER) >> tk(CRLF)
                         >> +(
                            !(( /* empty line is ok in typedef */
                                tk(IDENTIFIER) >> tk(AS) >> type
                                    /* { define it
                                       }
                                   */
                            ) % tk(','))
                            >> tk(CRLF)
                          )
                         >> tk(ENDTYPE) >> tk(CRLF)),
                        case_p<CONST> (
                            tk(IDENTIFIER) >> tk('=') >> expr
                         >> tk(CRLF)),
                        case_p<DATA> (
                            /* note: DATA must be at global level; not in SUBs etc. */
                            !( !tk(STRINGCONST) % tk(',') )
                         >> separator),
                        case_p<OPTION_BASE> (
                            tk(NUMBER) >> separator)
                    ]
                |   longstatements0
                    /* add eps_p here if it doesn't meddle with tk(EOF) detection */
                ),

                sub_params_decl =
                    tk('(') >> ((!(tk(BYVAL) | tk(SEG))
                           >> tk(IDENTIFIER) >> if_p(tk(AS)) [ type ]
                             /* if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; } */
                           ) % tk(','))
                 >> tk(')'),
                
                separator = tk(COLONCHAR) | tk(CRLF),

                type = tk(INTEGER)
                    |  tk(SINGLE)
                    |  tk(DOUBLE)
                    |  tk(STRING)
                    |  tk(LONG)
                    |  tk(IDENTIFIER)
                        /*
                            if(!IsType(*$1)) { YYABORT; }
                        */,

                /**** LVAL */
                var_target =
                    (
                        tk(IDENTIFIER)
                        /* perhaps with indexes */
                        >> if_p(tk('(')) [ (num_expr % tk(',')) >> tk(')') ]
                    ) % tk('.'),
                      /* expr_var* parent = dynamic_cast<expr_var*>($1);
                         expr_var* member = dynamic_cast<expr_var*>($3);
                         parent->set_ptr(member); $3 = 0; $$ = $1;
                       */
                
                autofilenum = tk(FILENUM) | num_expr,
                
                anylabeltarget =
                    tk(NUMBER)
                |   tk(IDENTIFIER),
                    /* automatically define as label */
                    
                printspec =
                (
                    tk(USING) >> string_expr >> tk(';')
                    >> (expr % (tk(',') | tk(';')))
                       /* params for format, delimiter identity=insignifncant */
                    >> !print_cursor_ctrl /* a print_cursor_ctrl may be given last */
                )
                | do_p[ expr | epsilon_p ].while_p(print_cursor_ctrl),
                        /* two exprs in row is invalid, must be delimited by cursor_ctrl */
                        /* many cursor_ctrls in row is valid */
                        /* whether the call ends with cursor_ctrl, is important */
                
                print_cursor_ctrl = tk(',') | tk(';'),

                plain_coordinate = tk('(') >> num_expr >> tk(',') >> num_expr >> tk(')'),
                step_coordinate = !tk(STEP) >> plain_coordinate,

                longstatements0 =
                *(  (   tk(IF) >> bool_expr >> tk(THEN) >> tk(CRLF) >> longstatements0
                           >> while_p(tk(ELSEIF))
                                [ bool_expr >> tk(THEN) >> longstatements0 ]
                           >> if_p(tk(ELSE)) [ longstatements0 ]
                           >> tk(ENDIF)
                    |   tk(FOR) >> var_target >> tk('=') >> num_expr >> tk(TO) >> num_expr
                            >> if_p(tk(STEP)) [ num_expr ]
                            >> separator
                            >> longstatements0
                            >> tk(NEXT)
                    |   tk(SELECT) >> tk(CASE)
                            >> expr >> +separator
                            >> while_p(tk(CASE))
                                [ casematch >> +separator >> longstatements0 ]
                            >> tk(ENDSELECT)
                    |   (   tk(DOWHILE)
                          | tk(DOUNTIL)
                        )        >> bool_expr >> separator
                                 >> longstatements0
                                 >> tk(LOOP)
                        |   tk(DO) >> separator
                                 >> longstatements0
                                 >> ( tk(LOOP)
                                    | tk(LOOPWHILE) >> bool_expr
                                    | tk(LOOPUNTIL) >> bool_expr
                                    )
                         /* do while ... loop
                            do until ... loop
                            do ... loop
                            do ... loop while
                            do ... loop until */
                    |   simplestatement
                  ) >> separator
                |   /* a label statement */
                    (   tk(NUMBER)
                    |   tk(IDENTIFIER) >> tk(COLONCHAR)
                    )
                |   shortif >> tk(CRLF)
                |   separator),
                
                /* shortstatements0 does not include trailing COLONCHARs. */
                shortstatements0 =
                    *tk(COLONCHAR)
                >>  /* statements that can be included between online THEN and ELSE */
                    /* Does not include shortif. */
                  !(
                    (   simplestatement
                    |   (   tk(DOWHILE)
                          | tk(DOUNTIL)
                        )        >> bool_expr >> tk(COLONCHAR)
                                 >> shortstatements0 >> *tk(COLONCHAR)
                                 >> tk(LOOP)
                        |   tk(DO) >> tk(COLONCHAR)
                                 >> shortstatements0 >> *tk(COLONCHAR)
                                 >> ( tk(LOOP)
                                    | tk(LOOPWHILE) >> bool_expr
                                    | tk(LOOPUNTIL) >> bool_expr
                                    )
                    |   tk(FOR) >> var_target >> tk('=') >> num_expr >> tk(TO) >> num_expr
                            >> if_p(tk(STEP)) [ num_expr ]
                            >> tk(COLONCHAR)
                            >> shortstatements0 >> *tk(COLONCHAR)
                            >> tk(NEXT)
                    |   tk(SELECT) >> tk(CASE) >> expr
                         >> +tk(COLONCHAR)
                         >> while_p(tk(CASE))
                             [ casematch >> +tk(COLONCHAR)
                            >> shortstatements0 >> +tk(COLONCHAR) ]
                         >> tk(ENDSELECT)
                    ) % +tk(COLONCHAR)
                   ),
                
                shortif =
                    tk(IF) >> bool_expr
                    >> tk(THEN) >> shortif_meat
                    >> *tk(COLONCHAR)
                    >> if_p(tk(ELSE)) [ shortif_meat ]
                    >> *tk(COLONCHAR),
                
                shortif_meat =
                    tk(NUMBER)
                |   shortstatements0
                    >> !(+tk(COLONCHAR) >> shortif),
                
                casematch =
                    (
                        tk(ELSE)                    /* $$ = new casevalues(ELSE);   */
                    |   tk(IS) >> switch_p
                          [
                              case_p<'>'>           ( expr ),
                              case_p<'<'>           ( expr ),
                              case_p<'='>           ( expr ),
                              case_p<LESSOREQUAL>   ( expr ),
                              case_p<GREATEROREQUAL>( expr ),
                              case_p<NOTEQUAL>      ( expr )
                          ]                           /* $$ = new casevalues({operator}, $3) */
                    |	expr >> if_p(tk(TO)) [ expr ] /* $$ = new casevalues($1, $3) */
                    ) % tk(','),

                dim_params =
                    /* DIM and REDIM.
                     * Note: REDIM can only be applied to $DYNAMIC arrays,
                     *             not scalars, utypes or $STATIC arrays.
                     */
                    !tk(SHARED) /* optional word */
                    >>  ((
                            tk(IDENTIFIER)
                            /* if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; } */
                         >> if_p(tk('(')) /* array dimension */
                            [ (( num_expr >> if_p(tk(TO)) [ num_expr ]
                               ) % tk(',')) >> tk(')') ]
                         >> if_p(tk(AS)) [ type ]
                         ) % tk(',')),
                
                simplestatement =
                 	switch_p
                    [
                        /* data commands */
                        
                        case_p<RESTORE>(
                            !anylabeltarget),
                        case_p<READ>(
                            var_target % tk(',')
                                    ),
                       
                        case_p<DIM> (dim_params),
                        case_p<REDIM> (dim_params),
                        case_p<ERASE> (
                                /* For static arrays, ERASE clears their contents to zero.
                                 * For dynamic arrays, ERASE shrinks them to zero. */
                                var_target % tk(',')
                            ),
                        
                        /* system commands */
                        case_p<SYSTEM>(
                            !string_expr),
                        case_p<POKE>(
                            num_expr >> tk(',') >> num_expr),
                        case_p<OUT>(
                            num_expr >> tk(',') >> num_expr),
                        case_p<WAIT>(
                            num_expr
                         >> tk(',')
                         >> num_expr
                         >> if_p(tk(',')) [ num_expr ] ),
                        case_p<DEFSEG>(
                            if_p(tk('=')) [ num_expr ] ),
                        
                        /* control commands */
                        
                        case_p<EXITDO>( epsilon_p ),
                        case_p<EXITFOR>( epsilon_p ),
                        case_p<EXITSUB>( epsilon_p ),
                        case_p<EXITFUNCTION>( epsilon_p ),
                        case_p<STOP>( epsilon_p ),
                        case_p<END>( epsilon_p ),
                        case_p<RESUME>(
                            tk(NEXT)
                        |   anylabeltarget
                        |   epsilon_p),
                        case_p<RETURN>(
                            !anylabeltarget),
                        case_p<GOTO>(
                            anylabeltarget),
                        case_p<GOSUB>(
                            anylabeltarget),

                        /* graphics commands */
                        
                        case_p<LOCATE>(
                            !num_expr >> repeat_p(0,4) [tk(',') >> !num_expr ]
                            /* row, col, cursor_vis, cursor_start, cursor_stop */
                            ), /* cursor_end */
                        case_p<SCREEN>(
                            !num_expr >> repeat_p(0,3) [tk(',') >> !num_expr ]
                            /* mode, colorswitch, apage, vpage */
                            ),
                        case_p<PAINT>(
                               step_coordinate
                            >> !(tk(',')
                            >> !string_or_num_expr /* pattern/paint */
                            >> !(tk(',')
                            >> !num_expr /* bordercolor */
                            >> !(tk(',')
                            >> !string_expr)))), /* background */
                        case_p<COLOR>(
                            !num_expr >> repeat_p(0,2) [tk(',') >> !num_expr ]
                            /* fgcolor, bgcolor, border */
                        ),
                        case_p<CIRCLE>(
                                 step_coordinate
                              >> tk(',') >> num_expr /* radius */
                              >> repeat_p(0,4) [tk(',') >> !num_expr]
                                /* color, start, end, aspect */
                        ),
                        case_p<GET>(
                               step_coordinate >> tk('-') >> step_coordinate
                            >> tk(',') >> var_target),
                        case_p<PUT>(
                               step_coordinate
                            >> tk(',') >> var_target
                            >> if_p(tk(','))
                                [ tk(XOR) | tk(PSET) | tk(PRESET) | tk(AND) | tk(OR) ]),
                                /* tk(XOR) is default */
                        case_p<VIEWPRINT>(
                                !(num_expr >> tk(TO) >> num_expr)),
                        case_p<VIEW>(
                                !tk(SCREEN)
                             >> plain_coordinate >> tk('-') >> plain_coordinate
                             >> !(tk(',')
                             >> !num_expr /* color */
                             >> !(tk(',')
                             >> !num_expr))), /* border */
                        case_p<PALETTE>(
                                tk(USING) >> var_target
                            |   !(num_expr >> tk(',') >> num_expr)),
                        case_p<PSET>(
                                step_coordinate
                             >> !(tk(',')
                             >> !num_expr)), /* color */
                        case_p<PRESET>(
                                step_coordinate
                               >> !(tk(',')
                               >> !num_expr)), /* color */
                        case_p<LINE>(
                                !step_coordinate >> tk('-') >> step_coordinate
                             >> !(tk(',')
                             >> !num_expr /* color */
                             >> !(tk(',')
                             >> !tk(IDENTIFIER) /* tk(B) | tk(BF) */
                             >> !(tk(',')
                             >> !num_expr)))), /* style (16-bit int mask) */
                        case_p<DRAW>(
                                string_expr),
                        case_p<CLS>(
                                !num_expr),

                        /* I/O commands */
                        
                        case_p<OPEN>(
                            string_expr /* filename */
                             >> if_p(tk(FOR))
                                 [ tk(RANDOM) | tk(OUTPUT) | tk(INPUT) | tk(APPEND) | tk(BINARY) ]
                                 /* RANDOM is default */
                             >> if_p(tk(ACCESS))
                                 [ tk(READ) || tk(WRITE) ]
                             >> !(tk(SHARED)
                               |  tk(LOCK) >> (tk(READ) || tk(WRITE)))
                             >> tk(AS) >> autofilenum
                             >> if_p(tk(LEN)) [ tk('=') >> num_expr ]
                        |   string_expr /* mode, first char is O/I/R/B/A */
                             >> tk(',') >>autofilenum
                             >> string_expr /* filename */
                             >> if_p(tk(',')) [ num_expr ]), /* reclen */
                        case_p<CLOSE>(
                            !(autofilenum % tk(','))), /* empty list = close all files */
                        case_p<PRINT>(
                            printspec),
                        case_p<PRINT_F>(
                            autofilenum >> tk(',') >> printspec),
                        case_p<LPRINT>(
                            printspec),
                        
                        /* event commands */
                        case_p<PLAY> (
                             switch_p
                             [
                                case_p<STOP> ( epsilon_p ),
                                case_p<OFF>  ( epsilon_p ),
                                case_p<ON>   ( epsilon_p )
                             ]
                           | (string_expr % tk(','))
                             ),
                        case_p<TIMER> (
                             switch_p
                             [
                                case_p<STOP> ( epsilon_p ),
                                case_p<OFF>  ( epsilon_p ),
                                case_p<ON>   ( epsilon_p )
                             ]),
                        case_p<KEY> (
                             switch_p
                             [
                                case_p<LIST> ( epsilon_p ),
                                case_p<OFF>  ( epsilon_p ),
                                case_p<ON>   ( epsilon_p )
                             ]
                           |  num_expr >> tk(',') >> string_expr
                             ),
                        case_p<ON_PLAY> (
                            tk('(') >> num_expr >> tk(')') >> tk(GOSUB) >> anylabeltarget
                                ),
                        case_p<ON_KEY> (
                            tk('(') >> num_expr >> tk(')') >> tk(GOSUB) >> anylabeltarget
                                ),
                        case_p<ON_TIMER> (
                            tk('(') >> num_expr >> tk(')') >> tk(GOSUB) >> anylabeltarget
                                ),
                        case_p<ON_ERROR> (
                            tk(GOTO) >> anylabeltarget
                                ),
                        case_p<ON> (
                            num_expr >> tk(GOSUB) >> (anylabeltarget % tk(','))
                        |   num_expr >> tk(GOTO) >> (anylabeltarget % tk(','))
                            )
                    ]
                
                |   !tk(LET) >> var_target >> tk('=') >> expr
                |	/* explicit sub call */
                    tk(CALL)
                    >> tk(IDENTIFIER) >> !('(' >> (expr % tk(',')) >> tk(')'))
                    /*
                        if(IsDefinedAsNonSub(*$2)) { YYABORT; }
                    */
                |   /* implicit sub call */
                    tk(IDENTIFIER) >> !(expr % tk(','))  /* Calling a SUB */
                    /*
                        if(IsDefinedAsNonSub(*$1)) { YYABORT; }
                    */
                    ,
            
                expr = QB_rvalue_p
            );
        }

        subrule<0 > program;
        subrule<1 > sub_params_decl;
        subrule<2 > separator;
        subrule<3 > type;
        subrule<4 > var_target;
        subrule<5 > autofilenum;
        subrule<6 > anylabeltarget;
        subrule<7 > printspec;
        subrule<8 > print_cursor_ctrl;
        subrule<9 > plain_coordinate;
        subrule<10> step_coordinate;
        
        subrule<11> longstatements0;
        subrule<12> shortstatements0;
        subrule<13> shortif;
        subrule<14> shortif_meat;
        subrule<15> casematch;
        subrule<16> dim_params;
        subrule<17> simplestatement;
        
        subrule<18> expr;
        /* aliases to expr: */
        subrule<18> num_expr;
        subrule<18> bool_expr;
        subrule<18> string_expr;
        subrule<18> string_or_num_expr;
        subrule<18> dummy_expr; /* discardable value */
        
        rule<ScannerT> first;
        
        const rule<ScannerT>& start() const { return first; }
    };
};

/************************************/

int main(void)
{
    /*setbuf(stdin,NULL);*/
    
    typedef multi_pass<int,
               multi_pass_policies::lex_input,
               multi_pass_policies::first_owner, 
               multi_pass_policies::no_check,
               multi_pass_policies::std_deque> iterator_t;
    typedef scanner<iterator_t> scanner_t;

    yyStartFile(stdin, "<stdin>");
    
    iterator_t begin( lex_get_token() );
    QB_Parser parser;
    parse(begin, iterator_t(0), parser);

    return 0;
}
